module reader
  use param, only: mline
  implicit none
  private

  public :: readfile
  public :: molecule, ifile_xyz, ifile_wfn, ifile_grd, averagenci
  public :: grid1, readgrid

  !> Radial grid type.
  type grid1
     logical :: init !< Is initialized?
     real*8 :: a !< Logarithmic grid parameter ri = a * exp(b * (i-1))
     real*8 :: b !< Logarithmic grid parameter ri = a * exp(b * (i-1))
     real*8 :: rmax !< Max. grid distance
     real*8 :: rmax2 !< Squared max. grid distance
     integer :: ngrid !< Number of nodes
     real*8, allocatable :: r(:) !< Node positions
     real*8, allocatable :: f(:) !< Grid values, f = 4*pi*r^2*rho
     real*8, allocatable :: fp(:) !< First derivative of f
     real*8, allocatable :: fpp(:) !< Second derivative of f 
  end type grid1

  !> Molecule type
  type molecule
     ! file name
     character(mline) :: name
     ! common for xyz and wfn
     integer :: ifile
     integer :: n
     real*8, allocatable :: x(:,:)
     integer, allocatable :: z(:)
     ! only for wfn
     integer :: nmo, npri
     integer, allocatable :: icenter(:)
     integer, allocatable :: itype(:)
     real*8, allocatable :: e(:)
     real*8, allocatable :: occ(:)
     real*8, allocatable :: c(:,:)
  end type molecule
  integer, parameter :: ifile_xyz = 1
  integer, parameter :: ifile_grd = 2
  integer, parameter :: ifile_wfn = 3

  !> Average NCI Information
  type averagenci
     ! average scheme: 1, first calcuate then average; 2, first average then calculate
!     integer :: scheme 

     ! frame # and weight
     integer :: nframe
     real*8  :: fweight

     ! other information
     logical :: ligand
     logical :: lig_in_all
     integer :: udat0

     ! saved info for each snap's S and rho, and coordinate
!     logical :: do_cpt
!     character*(mline) :: cptfile

!     integer :: nowlayer
!     integer :: scount
!     real*8, allocatable, dimension(:,:,:,:) :: srho, sdimgrad, isinter

!     integer :: cptcount
!     real*8, allocatable, dimension(:,:) :: criticalpt

     ! same as whole rhoparam
     real*8 :: rhoparam
 
     ! save nstep
     real*8 :: nstep(3)
     real*8 :: xinc(3)
     real*8 :: xinit(3)

     ! to estimate the fluctuation of environment, we need to save the density, gradient standard deviation
     logical :: do_std
     real*8, allocatable, dimension(:,:,:) :: std_rho, std_grad, std_rdg
     real*8  :: tmp_std_rho, tmp_std_grad

  end type averagenci

contains

  function readfile(file) result(m)
    use param
    use tools_io
    implicit none
    
    character*(mline), intent(in) :: file
    type(molecule) :: m

    character*(mline) :: ext

    ext = file(index(file,'.',.true.)+1:)
    call upper(ext)
    select case(adjustl(trim(ext)))
    case('XYZ')
       m = readxyz(file)
    case('WFN')
       m = readwfn(file)
    case default
       call error('readfile','Not recognized molecular format',faterr)
    end select

  end function readfile

  !> Read xyz file
  function readxyz(file) result(m)
    use param
    use tools_io
    implicit none

    character*(mline), intent(in) :: file
    type(molecule) :: m

    integer :: istat, i
    integer, parameter :: lu = 10
    character(mline) :: name

    ! set type
    m%ifile = ifile_xyz
    m%name = file

    ! read number of atoms
    open(lu,file=file,status='old')
    read(lu,*) m%n

    ! allocate arrays
    allocate(m%x(3,m%n),stat=istat)
    if (istat /= 0) call error('readxyz','could not allocate memory for atomic positions',faterr)
    allocate(m%z(m%n),stat=istat)
    if (istat /= 0) call error('readxyz','could not allocate memory for atomic numbers',faterr)

    ! read atomic coordinates and atomic numbers
    read(lu,*)
    do i = 1, m%n
       read (lu,*) name, m%x(:,i)
       m%z(i) = zatguess(name)
       if (m%z(i) == -1) call error('readxyz','atom type not recognized: '//trim(name),faterr)
       m%x(:,i) = m%x(:,i) / bohrtoa
       ! convert coordinate from Angstrom into Bohr
    end do
    
    ! close
    close(lu)

  end function readxyz

  !> Read wfn file
  function readwfn(file) result(m)
    use param
    use tools_io
    implicit none

    character*(mline), intent(in) :: file
    type(molecule) :: m

    integer, parameter :: lu = 10

    character*4 :: orbtyp
    integer :: i, j, istat
    real*8 :: zreal

    ! set type
    m%ifile = ifile_wfn
    m%name = file

    ! read number of atoms, primitives, orbitals
    open(lu,file=file,status='old')
    read (lu,*)
    read (lu,101) orbtyp, m%nmo, m%npri, m%n

    ! atomic positions and numbers
    allocate(m%x(3,m%n),stat=istat)
    if (istat /= 0) call error('readwfn','could not allocate memory for atomic positions',faterr)
    allocate(m%z(m%n),stat=istat)
    if (istat /= 0) call error('readwfn','could not allocate memory for atomic numbers',faterr)
    do i = 1, m%n
       read(lu,106) m%x(:,i), zreal
       m%z(i) = nint(zreal)
    end do

    ! center assignments, types of primitives
    allocate(m%icenter(m%npri),stat=istat)
    if (istat /= 0) call error('readwfn','could not allocate memory for icenter',faterr)
    allocate(m%itype(m%npri),stat=istat)
    if (istat /= 0) call error('readwfn','could not allocate memory for itype',faterr)
    read(lu,102) (m%icenter(i),i=1,m%npri)
    read(lu,102) (m%itype(i),i=1,m%npri)

    ! primitive exponents
    allocate(m%e(m%npri),stat=istat)
    if (istat /= 0) call error('readwfn','could not allocate memory for exponents',faterr)
    read(lu,103) (m%e(i),i=1,m%npri)
    
    ! occupations and orbital coefficients
    allocate(m%occ(m%nmo),stat=istat)
    if (istat /= 0) call error('readwfn','could not allocate memory for occupations',faterr)
    allocate(m%c(m%nmo,m%npri),stat=istat)
    if (istat /= 0) call error('readwfn','could not allocate memory for orbital coefficients',faterr)
    do i = 1, m%nmo
       read(lu,104) m%occ(i)
       read(lu,105) (m%c(i,j),j=1,m%npri)
    end do

    close(lu)

101 format (4X,A4,10X,3(I5,15X))
102 format(20X,20I3)
103 format(10X,5E14.7)
104 format(35X,F12.8)
105 format(5(E16.8))
106 format(24X,3(F12.8),10X,F5.1)

  end function readwfn

  !> Read a density core file from the database. 
  subroutine readgrid(g,z)
    use tools_io
    use param

    type(grid1), intent(out) :: g !< Output radial grid
    integer, intent(in) :: z !< Atomic number

    integer :: i, j
    real*8 :: xmin, zz, dx, r, r1, r2, r3 ,r4, delta, delta2
    integer :: ngrid, ns, ic
    integer, allocatable :: occ(:)
    real*8, allocatable :: wfcin(:), rr(:,:)
    character*2, allocatable :: wfcl(:)
    integer :: nn
    logical :: exist
    character*(mline) :: file

    integer, parameter :: lu = 10
    real*8, parameter :: core_cutdens = 1d-12 !< Cutoff contribution for core radial grids
    ! radial grid derivation formulas
    integer, parameter :: noef(6,3) = reshape((/&
       0,  1,  2,  3,  4,  5,&
       -2, -1,  0,  1,  2,  3,&
       -5, -4, -3, -2, -1,  0/),shape(noef)) !< Node offsets for 6-point derivation formulas.
    real*8, parameter :: coef1(6,3) = reshape((/&
       -274,  600, -600,  400,  -150,  24,&
       6,  -60,  -40,  120,   -30,   4,&
       -24, 150, -400,  600,  -600, 274 /),shape(coef1)) !< Coefficients for first derivative.
    real*8, parameter :: coef2(6,3) = reshape((/&
       225, -770,  1070,  -780,   305,   -50,&
       -5,   80,  -150,    80,    -5,     0,&
       -50,  305,  -780 ,  1070,  -770,   225/),shape(coef2)) !< Coefficients for second derivative.
    real*8, parameter :: fac1=1d0/120d0 !< Prefactor for first derivative.
    real*8, parameter :: fac2=2d0/120d0 !< Prefactor for second derivative.

    ! the filename
    file = trim(nciplot_dat) // nameguess(z) // "_lda" // ".wfc"

    ! check that the file exists
    inquire(file=file,exist=exist)
    if (.not.exist) then
       write (uout,'("File: ",A)') trim(file)
       call error("readgrid","Atomic density file not found",warning)
       g%init = .false.
       return
    end if

    ! Read header and allocate arrays
    open (unit=lu,file=file,status='old')
    read (lu,*) nn
    
    allocate(wfcin(nn),wfcl(nn),occ(nn))
    occ = 0d0
    read (lu,*) (wfcl(i),i=1,nn)
    read (lu,*) (occ(i),i=1,nn)
    read (lu,*) xmin, zz, dx, ngrid

    if (sum(occ) /= z) then
       ns = 0
       do i = 1, nn
          if (ns + occ(i) > z) then
             occ(i) = z - ns
             occ(i+1:nn) = 0d0
             exit
          else
             ns = ns + occ(i)
          end if
       end do
    end if

    ! Read the grid and build the density
    allocate(g%r(ngrid),rr(ngrid,0:2))
    rr = 0d0
    do i = 1, ngrid
       read (lu,*) r, (wfcin(j),j=1,nn)
       g%r(i) = r
       rr(i,0) = dot_product(occ(1:nn),wfcin(1:nn)**2)
       if (rr(i,0)/(4d0*pi*r**2) < core_cutdens .and. i > 1) then
          ngrid = i
          call realloc(g%r,ngrid)
          exit
       end if
    end do

    ! fill grid info
    g%init = .true.
    g%a = exp(xmin) / zz
    g%b = dx
    g%ngrid = ngrid
    g%rmax = g%r(ngrid)
    g%rmax2 = g%r(ngrid)**2

    ! calculate derivatives
    allocate(g%f(ngrid),g%fp(ngrid),g%fpp(ngrid))
    do i = 1, ngrid
       if (i <= 2) then
          ic = 1
       else if (i >= ngrid-2) then
          ic = 3
       else
          ic = 2
       end if
       do j = 1, 6
          rr(i,1) = rr(i,1) + coef1(j,ic) * rr(i+noef(j,ic),0)
          rr(i,2) = rr(i,2) + coef2(j,ic) * rr(i+noef(j,ic),0)
       end do
       rr(i,1) = rr(i,1) * fac1
       rr(i,2) = rr(i,2) * fac2

       r = g%r(i)
       r1 = 1d0 / r
       r2 = r1 * r1
       r3 = r2 * r1
       r4 = r3 * r1
       delta=1.d0/g%b
       delta2=delta*delta
       
       g%f(i) = rr(i,0) * r2
       g%fp(i) = (rr(i,1) * delta - 2.d0 * rr(i,0)) * r3
       g%fpp(i) = (rr(i,2) * delta2 - 5.d0 * rr(i,1) * delta + 6.d0 * rr(i,0)) * r4
    end do
    g%f = g%f / (4d0*pi)
    g%fp = g%fp / (4d0*pi)
    g%fpp = g%fpp / (4d0*pi)

    ! close the density file
    close(lu)

    ! cleanup
    deallocate(rr,wfcin,wfcl,occ)

  end subroutine readgrid

end module reader
